{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE DeriveTraversable #-}
module Sudoku.Area (Area(..), indexFromDigit, digitFromIndex) where

import Sudoku.Triple (Triple, Index)
import Data.Functor.Compose (Compose(..))
import Data.Functor.Rep (Representable)
import Data.Distributive (Distributive (..))
import Data.Bifunctor (Bifunctor(..))
import Control.Monad (join)

newtype Area a = Area { get :: Compose Triple Triple a }
  deriving stock (Foldable, Traversable)
  deriving newtype (Functor, Representable, Show)

instance Distributive Area where
  distribute :: Functor f => f (Area a) -> Area (f a)
  distribute = Area . distribute . fmap get

-- | Accepts arguments from `0` to `8`. Everything else fails with 'error'
--
-- prop> \ i -> digitFromIndex (indexFromDigit (i `mod` 9)) == i `mod` 9
-- +++ OK, passed 100 tests.
--
-- >>> indexFromDigit 9
-- *** Exception: Enum.toEnum{Index}: tag (3) is outside of enumeration's range (0,2)
--
-- >>> indexFromDigit 6
-- (Third,First)

indexFromDigit :: Int -> (Index, Index)
indexFromDigit i = join bimap toEnum $ i `divMod` 3

-- | Inverse of 'indexFromDigit'
-- 
-- >>> import Sudoku.Triple (Index(..))
-- >>> digitFromIndex (First, First)
-- 0

digitFromIndex :: (Index, Index) -> Int
digitFromIndex (outer, inner) = fromEnum outer * 3 + fromEnum inner

