{-# LANGUAGE LambdaCase #-}
module Sudoku.Render (unstructured, hints) where
import Sudoku ( Sudoku, orderedIndices, Sudoku(..) )
import Data.Functor.Compose (Compose(..))
import Sudoku.Field (Field)
import qualified Sudoku.Field as Field
import qualified Data.Functor.Rep as Representable
import Data.Set (Set)
import Graphics.Text.Box (Box, Border (..))
import qualified Graphics.Text.Box as Box
import qualified Data.Set as Set
import Data.Text (Text)
import Sudoku.Area (Area)
import Data.Bifunctor (bimap)
import Control.Monad (join)

unstructured :: Sudoku (Maybe Field) -> String
unstructured sudoku = let
     characters = Field.maybeToDigit <$> sudoku

     in unlines $ (Representable.index characters <$>) <$> orderedIndices

hints :: Sudoku (Either Field (Set Field)) -> Text
hints sudoku = let

  (Sudoku (Compose fieldBoxes)) = fieldBox <$> sudoku :: Sudoku Box
  areaBoxes = areaBox <$> fieldBoxes
  translateAreaIndices = join bimap toEnum
  boxStyle = Box.Border {verticalBetween=' ', topRight=' ', topLeft=' ', top=' ', right=' ', left=' ', horizontalBetween=' ', crosssection=' ', bottomRight=' ', bottomLeft=' ', bottom=' '}

  in Box.render $ Box.grid (Just boxStyle) (3, 3) (Representable.index areaBoxes . translateAreaIndices)

areaBox :: Area Box -> Box
areaBox area = let

  boxStyle = Box.asciiBorder {topLeft='#', topRight='#', bottomLeft='#', bottomRight='#'}
  translateIndices (i, j) = Representable.index area (toEnum i, toEnum j)

  in Box.grid (Just boxStyle) (3, 3) translateIndices

fieldBox :: Either Field (Set Field) -> Box
fieldBox = \case
  Left value -> Box.square 3 (const $ Field.toDigit value)
  Right options -> Box.square 3 $ \ (row, col) -> let
    option = toEnum $ row * 3 + col

    in if option `Set.member` options
      then Field.toDigit option
      else ' '

