{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE RequiredTypeArguments #-}
{-# LANGUAGE LambdaCase #-}
module Sudoku.Solve (solve, SolveError, hints) where
import Sudoku (Sudoku)
import Sudoku.Field (Field)
import qualified Sudoku.State as SudokuState
import Sudoku.State (SudokuState)
import qualified Data.Set as Set
import qualified Data.Functor.Representable as Representable
import Data.Set ( Set )
import Data.Functor.Rep (imapRep)

data SolveError
  = InvalidStartState
  deriving (Show)

-- | Given a state implementation and a sudoku. Calculate all the hints

hints :: forall s -> SudokuState s => Sudoku (Maybe Field) -> Sudoku (Either Field (Set Field))
hints s sudoku = let
  state = SudokuState.build @s sudoku

  getHint position = \case
    Just x  -> Left x -- field is already filled
    Nothing -> Right $ SudokuState.fieldOptions state position

  in imapRep getHint sudoku

-- | Given a State implementation and a sudoku field, try to solve it.

solve :: forall s -> SudokuState s => Sudoku (Maybe Field) -> Either SolveError [Sudoku Field]
solve s sudoku = let
  state = SudokuState.build @s sudoku
  in case SudokuState.isValid state of
    False -> Left InvalidStartState
    True -> Right $ solveWithState s sudoku state

solveWithState :: forall s -> SudokuState s => Sudoku (Maybe Field) -> s -> [Sudoku Field]
solveWithState s sudoku state = case sequence sudoku of
  Just solution -> [solution]
  Nothing -> fillField s sudoku state

fillField :: forall s -> SudokuState s => Sudoku (Maybe Field) -> s -> [Sudoku Field]
fillField s sudoku state = case SudokuState.nextField state of
  Nothing -> [] -- unsolvable
  Just position -> do
    -- list monad threads the execution
    fieldValue <- Set.toList (SudokuState.fieldOptions state position)

    -- update the stuff
    let newBoard = Representable.set position (Just fieldValue) sudoku
    let newState = SudokuState.setField state position fieldValue

    -- recurse
    solveWithState s newBoard newState

