{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE LambdaCase #-}
module Sudoku.Triple (Triple(..), Index(..), first, second, third) where
import Data.Distributive (Distributive (distribute, collect))
import Data.Functor.Rep
    ( Representable(Rep, tabulate, index), Co, Co(..) )

data Triple a = Triple a a a
  deriving stock (Show, Eq, Functor, Foldable, Traversable)
  deriving (Applicative, Monad) via Co Triple

first :: Triple a -> a
first  (Triple x _ _) = x
second :: Triple a -> a
second (Triple _ x _) = x
third :: Triple a -> a
third  (Triple _ _ x) = x

instance Distributive Triple where
  distribute :: Functor f => f (Triple a) -> Triple (f a)
  distribute x = Triple (fmap first x) (fmap second x) (fmap third x)
  collect :: Functor f => (a -> Triple b) -> f a -> Triple (f b)
  collect f x = distribute $ fmap f x

data Index = First | Second | Third
  deriving (Show, Eq, Ord, Enum, Bounded)

-- >>> toEnum 0 :: Index
-- First

instance Representable Triple where
  type Rep Triple = Index

  tabulate :: (Rep Triple -> a) -> Triple a
  tabulate f = Triple (f First) (f Second) (f Third)

  index :: Triple a -> Rep Triple -> a
  index (Triple x1 x2 x3) = \case
    First -> x1
    Second -> x2
    Third -> x3

-- >>> tabulate (const $ tabulate id) :: Triple (Triple Index)
-- Triple (Triple First Second Third) (Triple First Second Third) (Triple First Second Third)

